#!/usr/bin/python3
# coding: utf-8

"""
Slurm simulator's output trace multi-plot data script!

This script receives a file path as an argument and outputs, in the same folder
as where the input file is located, a bunch of plots:

- GLOBAL
 - Start job time histogram
 - Number of nodes histogram
 - Time in queue histogram
 - 2d histogram of time in queue X number of nodes
 - Submit job time histogram
 - Evolution of machine usage (REEEEALLY SLOW CUZ IT IS A FOR LOOP IN PYTHON)
 - Pie chart of the backfilled jobs
- PER USER
 - boxplot of start time job per user
 - boxplot of requested nodes per user
 - boxplot of wait time per user
 - Evolution of usage per user (REALLY SLOW because it is a for loop in python)

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2022
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt 
import sys  # To read input

#########################################
# HARDCODED MACHINE PARAMETERS!
TOTAL_NODES = 3456  # MN4
NO_CPUS_PER_NODE = 24
#########################################


def overall_start_time(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots histogram starting time and saves at dir_path.

    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting starting time histogram...")
    plt.hist(list(output_trace["Start time"]))
    plt.title(f"Distribution of Start time\n Trace: {file_name}")
    plt.xlabel("Time in seconds [since the start of the simulation(???)]")
    plt.savefig(f"{dir_path}/{file_name}_hist.png")
    plt.clf()


def overall_requested_nodes(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots histogram of requested number of nodes and saves at dir_path.

    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting requested nodes histogram...")
    plt.hist(list(output_trace["Nodes"]))
    plt.title(f"Distribution of Nodes requested\n Trace: {file_name}")
    plt.xlabel("Number of nodes")
    plt.savefig(f"{dir_path}/{file_name}_nodes_hist.png")
    plt.clf()


def overall_time_in_queue(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots histogram of time in queue and saves at dir_path.

    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting time in queue histogram...")
    plt.hist(list(output_trace["Wait time"]))
    plt.xlabel("Time in queue")
    plt.title(f"Time in Queue\n Trace: {file_name}")
    plt.savefig(f"{dir_path}/{file_name}_time_in_queue_hist.png")
    plt.clf()


def overall_2dhist_nodes_x_wait_time(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots 2D histogram of number of nodes X wait time for all jobs and saves at dir_path.

    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    # If FavorSmallerJobs = True, bigger jobs should wait longer in queue
    # If FavorSmallerJobs = False, smalled jobs should wait longer in queue
    print("Plotting 2d histogram (requested nodes) X (wait time)...")
    plt.hist2d(list(output_trace["Nodes"]), list(output_trace["Wait time"]))
    plt.xlabel("Number of Nodes")
    plt.ylabel("Time in Queue (seconds)")
    plt.title(f"Time in queue per number of nodes\n Trace: {file_name}")
    plt.savefig(f"{dir_path}/{file_name}_nodes_time_in_queue_hist2d.png")
    plt.clf()


def overall_nodes_x_wait_time(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
     Plots cloud of points of number of nodes X wait time for all jobs and saves at dir_path.

     :param output_trace: Pandas Data frame with output job traces
     :param file_name: name of the origin .csv file
     :param dir_path: path where the plot will be placed
     :return: None
     """
    # If FavorSmallerJobs = True, bigger jobs should wait longer in queue
    # If FavorSmallerJobs = False, smalled jobs should wait longer in queue
    print("Plotting (requested nodes) X (wait time)...")
    plt.plot(output_trace["Wait time"], output_trace["Nodes"], ".")
    plt.ylabel("Number of Nodes")
    plt.xlabel("Time in Queue (seconds)")
    plt.title(f"Number of nodes x Time in queue \n Trace: {file_name}")
    plt.savefig(f"{dir_path}/{file_name}_time_in_queue_nodes.png")
    plt.clf()

def overall_submit_time(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots histogram of submit time of all jobs and saves at dir_path
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting submit time histogram...")
    plt.hist(list(output_trace["Submit time"]))
    plt.title(f"Histogram of submit time\n Trace: {file_name}")
    plt.xlabel("Submit time in seconds [since the begining of simulation?]")
    plt.savefig(f"{dir_path}/{file_name}_submit_time_hist.png")
    plt.clf()


def overall_evolution_machine_usage(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots evolution of machine usage and saves at dir_path
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting instant total machine usage...")
    # start_time = np.min(output_trace["Start time"])
    final_time = np.max(output_trace["End time"])+1
    prop_nodes_per_time = []
    # we go through every second from start_time and final_time and compute the total number of active processes
    for i in range(final_time):
        active_jobs = output_trace[(output_trace["Start time"] < i) & (output_trace["End time"] >= i)]
        used_nodes = np.sum(active_jobs["Nodes"])
        prop_node = used_nodes/TOTAL_NODES  # proportion of used nodes in T+i seconds
        # we multiple by 100 to have percentage!
        prop_nodes_per_time.append(100*prop_node)

    plt.plot(range(final_time), prop_nodes_per_time)
    plt.title(f"Instant machine-wide total usage\n Trace: {file_name}")
    plt.xlabel("Time in seconds")
    plt.ylabel("% of Usage")
    plt.savefig(f"{dir_path}/{file_name}_instant_usage.png")
    plt.clf()

def overall_backfilled(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Pie chart of backfilled jobs
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting pie chart of the backfilled jobs...")
    labels = "backfilled", "not backfilled"
    no_backfilled = len(output_trace[(output_trace["Backfilled"] == 1)]) # number of backfilled jobs
    total = len(output_trace) # total number of jobs

    plt.pie([no_backfilled, total - no_backfilled], labels=labels)
    plt.title("Proportion of backfilled jobs")
    plt.savefig(f"{dir_path}/{file_name}_backfilled_proportion.png")
    plt.clf()

# manuel: Hardcoded UID and associations...
# users[1001] = "user1/account1"
# users[1002] = "user2/account2"
# users[1003] = "tester/account1"


def user_start_time(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots boxplot starting time per user and saves at dir_path
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    # starting time (this is kinda of stupid, isn't?)
    print("Plotting start time per user...")
    users = output_trace["UserId"].unique()
    start_time_per_user = []
    for user in users:
        start_time_per_user.append(output_trace[output_trace["UserId"] == user]["Start time"])
    plt.boxplot(start_time_per_user, labels=users)
    plt.title(f"Start time of jobs per user\n Trace: {file_name}")
    plt.ylabel("Time in seconds")
    plt.savefig(f"{dir_path}/{file_name}_start_time_boxplot.png")
    plt.clf()


def user_nodes_requested(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots boxplot of nodes requested per user and saves at dir_path
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting nodes requested per user...")
    # Find out all the users that sent jobs in this simulation
    users = output_trace["UserId"].unique()
    nodes_per_user = []
    for user in users:
        nodes_per_user.append(output_trace[output_trace["UserId"] == user]["Nodes"])
    plt.boxplot(nodes_per_user, labels=users)
    plt.title(f"Nodes requested per user\n Trace: {file_name}")
    plt.ylabel("Number of nodes")
    plt.savefig(f"{dir_path}/{file_name}_nodes_boxplot.png")
    plt.clf()


def user_wait_time(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots boxplot of wait time per user and saves at dir_path
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting wait time per user...")
    # Find out all the users that sent jobs in this simulation
    users = output_trace["UserId"].unique()
    wait_time_per_user = []
    for user in users:
        wait_time_per_user.append(output_trace[output_trace["UserId"] == user]["Wait time"])
    plt.boxplot(wait_time_per_user, labels=users)
    plt.title(f"Wait time of jobs per user\n Trace: {file_name}")
    plt.ylabel("Time in seconds")
    plt.savefig(f"{dir_path}/{file_name}_wait_time_boxplot.png")
    plt.clf()


def user_usage(output_trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Plots the evolution of usage for all users and outputs to dir_path
    :param output_trace: Pandas Data frame with output job traces
    :param file_name: name of the origin .csv file
    :param dir_path: path where the plot will be placed
    :return: None
    """
    print("Plotting usage per user...")
    users = output_trace["UserId"].unique()
    # Reference point is the first starting job
    # start_time = min(output_trace["Start time"])
    start_time = 0  # start at 0
    final_time = max(output_trace["End time"])+1
    # array of arrays. Each of the arrays is the usage of the user in second i-th second from reference point start_time
    usage_evolution = []
    # COMPUTE INSTANT USAGE FOR EACH USER
    for user in users:
        user_jobs = output_trace[output_trace["UserId"] == user]
        user_usage_evolution = []
        # go through every second of the simulation
        for t in range(start_time, final_time):
            finished_jobs = user_jobs[user_jobs["End time"] < t]
            current_jobs = user_jobs[(user_jobs["Start time"] < t) & (user_jobs["End time"] >= t)]
            # RUNTIME*CORES is usage
            finished_usage = np.sum(finished_jobs["Run time"]*finished_jobs["Nodes"]*NO_CPUS_PER_NODE)
            current_usage = np.sum((t-current_jobs["Start time"])*current_jobs["Nodes"]*NO_CPUS_PER_NODE)
            user_usage_evolution.append(finished_usage+current_usage)
        usage_evolution.append(user_usage_evolution)

    # PLOT FOR EACH USER USAGE
    for i in range(len(users)):
        plt.plot(range(final_time), usage_evolution[i], label=users[i])

    plt.xlabel("Time in seconds")
    plt.ylabel("Usage ($Cores \cdot s$)")
    plt.title(f"Evolution of usage per user\n Trace: {file_name}")
    plt.legend()
    plt.savefig(f"{dir_path}/{file_name}_instant_usage_per_user.png", dpi=150)
    plt.clf()


if __name__ == "__main__":
    if len(sys.argv) == 1:
        print("You have to provide a file to read")
        sys.exit(1)

    file_path = sys.argv[1]
    simulation_name = file_path.split("/")[-1].split(".")[0]  # pick the file name and remove its file type
    simulation_path = file_path.rsplit("/", 1)[0]
    print(f"Reading output trace from {file_path}...")
    out_trace = pd.read_csv(file_path, sep=";")

    # ALL PLOTS
    overall_start_time(out_trace, simulation_name, simulation_path)
    overall_requested_nodes(out_trace, simulation_name, simulation_path)
    overall_time_in_queue(out_trace, simulation_name, simulation_path)
    overall_2dhist_nodes_x_wait_time(out_trace, simulation_name, simulation_path)
    overall_nodes_x_wait_time(out_trace, simulation_name, simulation_path)
    overall_submit_time(out_trace, simulation_name, simulation_path)
    # overall_evolution_machine_usage(out_trace, simulation_name, simulation_path)
    overall_backfilled(out_trace, simulation_name, simulation_path)
    user_start_time(out_trace, simulation_name, simulation_path)
    user_nodes_requested(out_trace, simulation_name, simulation_path)
    user_wait_time(out_trace, simulation_name, simulation_path)
    # user_usage(out_trace, simulation_name, simulation_path)
